@@ -65,20 +65,50 @@ def _normalize_traits(traits: ty.List[str]) -> ty.List[str]:
     return list(standard_traits) + custom_traits
 
 
+def get_traits(traits_str: str) -> ty.Set[str]:
+    """Return a normalized set of placement standard and custom traits from
+    a string of comma separated trait names.
+    """
+    # traits is a comma separated list of placement trait names
+    if not traits_str:
+        return set()
+    return set(_normalize_traits(traits_str.split(',')))
+
+
 def _get_traits_for_dev(
     dev_spec_tags: ty.Dict[str, str],
 ) -> ty.Set[str]:
-    # traits is a comma separated list of placement trait names
-    traits_str = dev_spec_tags.get("traits")
-    if not traits_str:
-        return {os_traits.COMPUTE_MANAGED_PCI_DEVICE}
-
-    traits = traits_str.split(',')
-    return set(_normalize_traits(traits)) | {
+    return get_traits(dev_spec_tags.get("traits", "")) | {
         os_traits.COMPUTE_MANAGED_PCI_DEVICE
     }
 
 
+def _normalize_resource_class(rc: str) -> str:
+    rc = rc.upper()
+    if (
+            rc not in os_resource_classes.STANDARDS and
+            not os_resource_classes.is_custom(rc)
+    ):
+        rc = os_resource_classes.normalize_name(rc)
+        # mypy: normalize_name will return non None for non None input
+        assert rc
+
+    return rc
+
+
+def get_resource_class(
+    requested_name: ty.Optional[str], vendor_id: str, product_id: str
+) -> str:
+    """Return the normalized resource class name based on what is requested
+    or if nothing is requested then generated from the vendor_id and product_id
+    """
+    if requested_name:
+        rc = _normalize_resource_class(requested_name)
+    else:
+        rc = f"CUSTOM_PCI_{vendor_id}_{product_id}".upper()
+    return rc
+
+
 def _get_rc_for_dev(
     dev: pci_device.PciDevice,
     dev_spec_tags: ty.Dict[str, str],
@@ -91,23 +121,8 @@ def _get_rc_for_dev(
     The user specified resource class is normalized if it is not already an
     acceptable standard or custom resource class.
     """
-    # Either use the resource class from the config or the vendor_id and
-    # product_id of the device to generate the RC
     rc = dev_spec_tags.get("resource_class")
-    if rc:
-        rc = rc.upper()
-        if (
-            rc not in os_resource_classes.STANDARDS and
-            not os_resource_classes.is_custom(rc)
-        ):
-            rc = os_resource_classes.normalize_name(rc)
-            # mypy: normalize_name will return non None for non None input
-            assert rc
-
-    else:
-        rc = f"CUSTOM_PCI_{dev.vendor_id}_{dev.product_id}".upper()
-
-    return rc
+    return get_resource_class(rc, dev.vendor_id, dev.product_id)
 
 
 class PciResourceProvider:
@@ -119,14 +134,24 @@ class PciResourceProvider:
         self.children_devs: ty.List[pci_device.PciDevice] = []
         self.resource_class: ty.Optional[str] = None
         self.traits: ty.Optional[ty.Set[str]] = None
+        # This is an adjustment for the total inventory based on normal device
+        # due to possibility of devices held in the tracker even though they
+        # are removed from the configuration due to still having allocations.
+        # This number will be calculated based on the existing allocations
+        # during update_provider_tree call.
+        self.adjustment = 0
 
     @property
     def devs(self) -> ty.List[pci_device.PciDevice]:
         return [self.parent_dev] if self.parent_dev else self.children_devs
 
+    @property
+    def total(self):
+        return len(self.devs) + self.adjustment
+
     @property
     def to_be_deleted(self):
-        return not bool(self.devs)
+        return self.total == 0
 
     def add_child(self, dev, dev_spec_tags: ty.Dict[str, str]) -> None:
         if self.parent_dev:
@@ -200,22 +225,56 @@ class PciResourceProvider:
             ]
         )
 
+    def _adjust_for_removals_and_held_devices(
+        self,
+        provider_tree: provider_tree.ProviderTree,
+        rp_rc_usage: ty.Dict[str, ty.Dict[str, int]],
+    ) -> None:
+
+        rp_uuid = provider_tree.data(self.name).uuid
+        rc_usage = rp_rc_usage[rp_uuid]
+
+        if not self.resource_class:
+            # The resource_class is undefined when there are no normal devices
+            # exists any more on this RP. If no normal devs exists then there
+            # is no device_spec to derive the RC and traits from. But if we
+            # still have allocations in placement against this RP that means
+            # there are devices removed from the configuration but kept in the
+            # tracker as they are still allocated. In this case we
+            # need to recover the resource class and traits from the
+            # existing allocation.
+            if len(rc_usage) == 0:
+                # no usage so nothing to adjust here
+                return
+            else:
+                # The len > 1 case should not happen for PCI RPs as we either
+                # track the parent PF or the child VFs there on the RP but
+                # never both.
+                self.resource_class = list(rc_usage.keys())[0]
+                self.traits = provider_tree.data(rp_uuid).traits
+
+        # If device being removed but still held due to still having
+        # allocations then we need to adjust the total inventory to never go
+        # below the current usage otherwise Placement will reject the update.
+        usage = rc_usage[self.resource_class]
+        inventory = self.total
+        if usage > inventory:
+            LOG.warning(
+                "Needed to adjust inventories of %s on "
+                "resource provider %s from %d to %d due to existing "
+                "placement allocations. This should only happen while "
+                "VMs using already removed devices.",
+                self.resource_class, self.name, inventory, usage)
+            # This is counted into self.total to adjust the inventory
+            self.adjustment += usage - inventory
+
     def update_provider_tree(
         self,
         provider_tree: provider_tree.ProviderTree,
         parent_rp_name: str,
+        rp_rc_usage: ty.Dict[str, ty.Dict[str, int]],
     ) -> None:
 
-        if self.to_be_deleted:
-            # This means we need to delete the RP from placement if exists
-            if provider_tree.exists(self.name):
-                # NOTE(gibi): If there are allocations on this RP then
-                # Placement will reject the update the provider_tree is
-                # synced up.
-                provider_tree.remove(self.name)
-
-            return
-
         if not provider_tree.exists(self.name):
             # NOTE(gibi): We need to generate UUID for the new provider in Nova
             # instead of letting Placement assign one. We are potentially
@@ -228,6 +287,14 @@ class PciResourceProvider:
                 uuid=uuidutils.generate_uuid(dashed=True)
             )
 
+        self._adjust_for_removals_and_held_devices(provider_tree, rp_rc_usage)
+
+        # if after the adjustment no inventory left then we need to delete
+        # the RP explicitly
+        if self.total == 0:
+            provider_tree.remove(self.name)
+            return
+
         provider_tree.update_inventory(
             self.name,
             # NOTE(gibi): The rest of the inventory fields (reserved,
@@ -239,13 +306,19 @@ class PciResourceProvider:
             # value periodically.
             {
                 self.resource_class: {
-                    "total": len(self.devs),
-                    "max_unit": len(self.devs),
+                    "total": self.total,
+                    "max_unit": self.total,
                 }
             },
         )
         provider_tree.update_traits(self.name, self.traits)
 
+        # Here we are sure the RP exists in the provider_tree. So, we can
+        # record the RP UUID in each PciDevice this RP represents
+        rp_uuid = provider_tree.data(self.name).uuid
+        for dev in self.devs:
+            dev.extra_info['rp_uuid'] = rp_uuid
+
     def update_allocations(
         self,
         allocations: dict,
@@ -325,13 +398,13 @@ class PciResourceProvider:
         return updated
 
     def __str__(self) -> str:
-        if self.devs:
+        if not self.to_be_deleted:
             return (
-                f"RP({self.name}, {self.resource_class}={len(self.devs)}, "
+                f"RP({self.name}, {self.resource_class}={self.total}, "
                 f"traits={','.join(sorted(self.traits or set()))})"
             )
         else:
-            return f"RP({self.name}, <EMPTY>)"
+            return f"RP({self.name}, <to be deleted>)"
 
 
 class PlacementView:
@@ -365,18 +438,6 @@ class PlacementView:
 
         return self._get_rp_name_for_address(dev.parent_addr)
 
-    def _add_child(
-        self, dev: pci_device.PciDevice, dev_spec_tags: ty.Dict[str, str]
-    ) -> None:
-        rp_name = self._get_rp_name_for_child(dev)
-        self._ensure_rp(rp_name).add_child(dev, dev_spec_tags)
-
-    def _add_parent(
-        self, dev: pci_device.PciDevice, dev_spec_tags: ty.Dict[str, str]
-    ) -> None:
-        rp_name = self._get_rp_name_for_address(dev.address)
-        self._ensure_rp(rp_name).add_parent(dev, dev_spec_tags)
-
     def _add_dev(
         self, dev: pci_device.PciDevice, dev_spec_tags: ty.Dict[str, str]
     ) -> None:
@@ -387,10 +448,11 @@ class PlacementView:
             # devices in placement.
             return
 
+        rp = self._ensure_rp_for_dev(dev)
         if dev.dev_type in PARENT_TYPES:
-            self._add_parent(dev, dev_spec_tags)
+            rp.add_parent(dev, dev_spec_tags)
         elif dev.dev_type in CHILD_TYPES:
-            self._add_child(dev, dev_spec_tags)
+            rp.add_child(dev, dev_spec_tags)
         else:
             msg = _(
                 "Unhandled PCI device type %(type)s for %(dev)s. Please "
@@ -401,51 +463,92 @@ class PlacementView:
             }
             raise exception.PlacementPciException(error=msg)
 
-    def _remove_child(self, dev: pci_device.PciDevice) -> None:
-        rp_name = self._get_rp_name_for_child(dev)
-        self._ensure_rp(rp_name).remove_child(dev)
-
-    def _remove_parent(self, dev: pci_device.PciDevice) -> None:
-        rp_name = self._get_rp_name_for_address(dev.address)
-        self._ensure_rp(rp_name).remove_parent(dev)
-
     def _remove_dev(self, dev: pci_device.PciDevice) -> None:
         """Remove PCI devices from Placement that existed before but now
         deleted from the hypervisor or unlisted from [pci]device_spec
         """
+        rp = self._ensure_rp_for_dev(dev)
         if dev.dev_type in PARENT_TYPES:
-            self._remove_parent(dev)
+            rp.remove_parent(dev)
         elif dev.dev_type in CHILD_TYPES:
-            self._remove_child(dev)
+            rp.remove_child(dev)
+
+    def _ensure_rp_for_dev(
+        self, dev: pci_device.PciDevice
+    ) -> PciResourceProvider:
+        """Ensures that the RP exists for the device and returns it
+        but does not do any inventory accounting for the given device on
+        the RP.
+        """
+        if dev.dev_type in PARENT_TYPES:
+            rp_name = self._get_rp_name_for_address(dev.address)
+            return self._ensure_rp(rp_name)
+        elif dev.dev_type in CHILD_TYPES:
+            rp_name = self._get_rp_name_for_child(dev)
+            return self._ensure_rp(rp_name)
+        else:
+            raise ValueError(
+                f"Unhandled PCI device type {dev.dev_type} "
+                f"for dev {dev.address}.")
 
     def process_dev(
         self,
         dev: pci_device.PciDevice,
         dev_spec: ty.Optional[devspec.PciDeviceSpec],
     ) -> None:
+        # NOTE(gibi): We never observer dev.status DELETED as when that is set
+        # the device is also removed from the PCI tracker. So we can ignore
+        # that state.
+        if dev.status == fields.PciDeviceStatus.REMOVED:
+            # NOTE(gibi): We need to handle the situation when an instance
+            # uses a device where a dev_spec is removed. Here we need to keep
+            # the device in the Placement view similarly how the PCI tracker
+            # does it.
+            # However, we also need to handle the situation when such VM is
+            # being deleted. In that case we are called after the dev is freed
+            # and marked as removed by the tracker so dev.instance_uuid is
+            # None and dev.status is REMOVED. At this point the Placement
+            # allocation for this dev is still not deleted so we still have to
+            # keep the device in our view. The device will be deleted when the
+            # PCI tracker is saved which happens after us.
+            # However, we cannot overly eagerly keep devices here as a
+            # device in REMOVED state might be a device that had no allocation
+            # in Placement so it can be removed already without waiting for
+            # the next periodic update when the device disappears from the
+            # PCI tracker's list. If we are over eagerly keeping such device
+            # when it is not allocated then that will prevent a single step
+            # reconfiguration from whitelisting a VF to whitelisting its
+            # parent PF, because the VF will be kept at restart and conflict
+            # with the PF being added.
 
-        if dev.status in (
-            fields.PciDeviceStatus.DELETED,
-            fields.PciDeviceStatus.REMOVED,
-        ):
-            # If the PCI tracker marked the device DELETED or REMOVED then
-            # such device is not allocated, so we are free to drop it from
-            # placement too.
+            # We choose to remove these devs so the happy path of removing
+            # not allocated devs is simple. And then we do an extra
+            # step later in update_provider_tree to reconcile Placement
+            # allocations with our view and add back some inventories to handle
+            # removed but allocated devs.
             self._remove_dev(dev)
         else:
             if not dev_spec:
                 if dev.instance_uuid:
                     LOG.warning(
                         "Device spec is not found for device %s in "
-                        "[pci]device_spec. We are skipping this devices "
-                        "during Placement update. The device is allocated by "
-                        "%s. You should not remove an allocated device from "
-                        "the configuration. Please restore the configuration "
-                        "or cold migrate the instance to resolve the "
-                        "inconsistency.",
+                        "[pci]device_spec. The device is allocated by "
+                        "%s. We are keeping this device in the Placement "
+                        "view. You should not remove an allocated device from "
+                        "the configuration. Please restore the configuration. "
+                        "If you cannot restore the configuration as the "
+                        "device is dead then delete or cold migrate the "
+                        "instance and then restart the nova-compute service "
+                        "to resolve the inconsistency.",
                         dev.address,
                         dev.instance_uuid
                     )
+                    # We need to keep the RP, but we cannot just use _add_dev
+                    # to generate the inventory on the RP as that would require
+                    # to know the dev_spec to e.g. have the RC. So we only
+                    # ensure that the RP exists, the inventory will be adjusted
+                    # based on the existing allocation in a later step.
+                    self._ensure_rp_for_dev(dev)
                 else:
                     LOG.warning(
                         "Device spec is not found for device %s in "
@@ -465,11 +568,50 @@ class PlacementView:
             f"{', '.join(str(rp) for rp in self.rps.values())}"
         )
 
-    def update_provider_tree(
+    @staticmethod
+    def get_usage_per_rc_and_rp(
+        allocations
+    ) -> ty.Dict[str, ty.Dict[str, int]]:
+        """Returns a dict keyed by RP uuid and the value is a dict of
+        resource class: usage pairs telling how much total usage the given RP
+        has from the given resource class across all the allocations.
+        """
+        rp_rc_usage: ty.Dict[str, ty.Dict[str, int]] = (
+            collections.defaultdict(lambda: collections.defaultdict(int)))
+        for consumer in allocations.values():
+            for rp_uuid, alloc in consumer["allocations"].items():
+                for rc, amount in alloc["resources"].items():
+                    rp_rc_usage[rp_uuid][rc] += amount
+
+        return rp_rc_usage
+
+    def _remove_managed_rps_from_tree_not_in_view(
         self, provider_tree: provider_tree.ProviderTree
     ) -> None:
+        """Removes PCI RPs from the provider_tree that are not present in the
+        current PlacementView.
+        """
+        rp_names_in_view = {rp.name for rp in self.rps.values()}
+        uuids_in_tree = provider_tree.get_provider_uuids_in_tree(
+            self.root_rp_name)
+        for rp_uuid in uuids_in_tree:
+            rp_data = provider_tree.data(rp_uuid)
+            is_pci_rp = provider_tree.has_traits(
+                rp_uuid, [os_traits.COMPUTE_MANAGED_PCI_DEVICE])
+            if is_pci_rp and rp_data.name not in rp_names_in_view:
+                provider_tree.remove(rp_uuid)
+
+    def update_provider_tree(
+        self,
+        provider_tree: provider_tree.ProviderTree,
+        allocations: dict,
+    ) -> None:
+        self._remove_managed_rps_from_tree_not_in_view(provider_tree)
+
+        rp_rc_usage = self.get_usage_per_rc_and_rp(allocations)
         for rp_name, rp in self.rps.items():
-            rp.update_provider_tree(provider_tree, self.root_rp_name)
+            rp.update_provider_tree(
+                provider_tree, self.root_rp_name, rp_rc_usage)
 
     def update_allocations(
         self,
@@ -579,16 +721,22 @@ def update_provider_tree_for_pci(
         dev_spec = pci_tracker.dev_filter.get_devspec(dev)
         pv.process_dev(dev, dev_spec)
 
+    pv.update_provider_tree(provider_tree, allocations)
+
     LOG.info("Placement PCI resource view: %s", pv)
 
-    pv.update_provider_tree(provider_tree)
     old_alloc = copy.deepcopy(allocations)
+    # update_provider_tree correlated the PciDevice objects with RPs in
+    # placement and recorded the RP UUID in the PciDevice object. We need to
+    # trigger an update on the device pools in the tracker to get the device
+    # RP UUID mapped to the device pools
+    pci_tracker.stats.populate_pools_metadata_from_assigned_devices()
     updated = pv.update_allocations(allocations, provider_tree)
 
     if updated:
         LOG.debug(
             "Placement PCI view needs allocation healing. This should only "
-            "happen if [scheduler]pci_in_placement is still disabled. "
+            "happen if [filter_scheduler]pci_in_placement is still disabled. "
             "Original allocations: %s New allocations: %s",
             old_alloc,
             allocations,
